/**
 * crype
 */

#include "ccrype.h"

#include "vendor\pbkdf2_sha1\sha1-pbkdf2.h"
#include "vendor\chacha\chacha.h"
#include "vendor\miniz\miniz.h"

// n.b. There is an updated miniz available in valvesoftware/vogl with bugfixes / Zip64 support / Zip archive
//   support made optional, but it does not build cleanly and requires some further work to separate from vogl stuff.
// @ref http://richg42.blogspot.co.nz/2014/03/zip64-version-of-miniz-library-released.html

#define CRYPE_PBKDF2_ITERATIONS 4096

// Chunks to read from stdin at once. We should probably keep this below linux PIPE_BUF (default 4096) since that
//  is the largest amount guaranteed to be able to read atomically (does that matter for us?)
#define CRYPE_STDIN_READ_SIZE 8192

// Trigger if the sum of the past {CRYPE_DECISION_YAW} bytes equals zero mod {CRYPE_DECISION_PITCH}.
#define CRYPE_DECISION_YAW 256
#define CRYPE_DECISION_PITCH 4096

#define CRYPE_BUFFER_RESERVE ( 256 * CRYPE_DECISION_PITCH )

#define CRYPE_MIN_SINGLE_LENGTH 4096
#define CRYPE_MAX_SINGLE_LENGTH ( 16 * 1024 * 1024 )

#define CRYPE_KEYSIZE 256

void crype::set_encryption_key(std::string& pass_str) {

	memset(enckey, 0, 32);

	// Derive 256-bit encryption key from passphrase into static global
	pbkdf2_sha1(pass_str.c_str(), pass_str.length(), CRYPE_PBKDF2_ITERATIONS, enckey, CRYPE_KEYSIZE/8);
}


void crype::handle_enc(std::vector<unsigned char>& buff) {

	// Chacha buffer
	ECRYPT_ctx ctx = { 0 };
	ECRYPT_keysetup(&ctx, enckey, CRYPE_KEYSIZE, 0); // zero-length IV/nonce not used

	std::vector<unsigned char> crypted(buff.size());
	ECRYPT_encrypt_bytes(&ctx, &buff[0], &crypted[0], buff.size());

	// Echo output
	fwrite(&crypted[0], crypted.size(), 1, out);

}


void crype::handle_zip_enc(std::vector<unsigned char>& buff) {
	if (!buff.size()) {
		return; // fancy hat and nothing to say
	}

	// Compress buffer
	std::vector<unsigned char> compressed;
	compressed.resize(buff.size());
	unsigned long compressed_size = compressed.size();
	int did_compress = mz_compress2(&compressed[0], &compressed_size, &buff[0], buff.size(), MZ_DEFAULT_COMPRESSION);

	// If compressed...
	if (did_compress == MZ_OK) {
		compressed.resize(compressed_size); // truncate

		unsigned int marker = compressed.size();
		marker |= (1 << 31); // is_compressed=true
		fwrite(&marker, sizeof(unsigned int), 1, out);

		handle_enc(compressed);

	} else {
		// compression failed, or didn't become any smaller than the input.
		// Just encrypt original data

		unsigned int marker = buff.size();
		fwrite(&marker, sizeof(unsigned int), 1, out);

		handle_enc(buff);

	}

	// Clear buff
	buff.clear();

	total_chains++;
}


bool crype::encrype() {

	unsigned int yaw_sum = 0;
	unsigned char yaw[CRYPE_DECISION_YAW] = { 0 };
	unsigned int yaw_pos = 0;

	std::vector<unsigned char> towrite;
	towrite.reserve(CRYPE_BUFFER_RESERVE);

	// .

	for (;;) {
		// Check for eof/ferror on input
		if (feof(in) || ferror(in)) {
			break;
		}

		// Read from input
		unsigned char buff[CRYPE_STDIN_READ_SIZE] = { 0 };
		size_t read = fread(buff, 1, CRYPE_STDIN_READ_SIZE, in);
		if (!read) continue;

		for (size_t i = 0; i < read; ++i) {

			// Copy byte
			towrite.push_back(buff[i]);

			// Rolling sum of last _YAW bytes
			unsigned char yaw_last = yaw[yaw_pos];
			yaw[yaw_pos] = buff[i];
			yaw_pos = (yaw_pos + 1) % CRYPE_DECISION_YAW;
			yaw_sum = (yaw_sum + buff[i] - yaw_last) % CRYPE_DECISION_PITCH; // This might overflow negative(?)

			// Take decision on plaintext input
			if (yaw_sum == 0 && towrite.size() > CRYPE_MIN_SINGLE_LENGTH) {
				handle_zip_enc(towrite);
			}

		}

		// Force a chain if we havn't had one in ages
		if (towrite.size() > CRYPE_MAX_SINGLE_LENGTH) {
			handle_zip_enc(towrite);
		}
	}

	// Flush remaining buffers
	handle_zip_enc(towrite);

	return true;
}


void crype::handle_unzip(std::vector<unsigned char>& chain) {

	// FIXME: BAD ESTIMATE and lots of memory usage
	std::vector<unsigned char> uncompressed(chain.size() * 1024);

	unsigned long uncompressed_len = uncompressed.size();
	int did_uncompress = mz_uncompress(&uncompressed[0], &uncompressed_len, &chain[0], chain.size());
	if (did_uncompress != MZ_OK) {
		fprintf(stderr, "crype: failed to decompress! (%d)\n", did_uncompress);
		exit(1);

	}

	// Echo output
	fwrite(&uncompressed[0], uncompressed_len, 1, out);

}


void crype::handle_chain_unzip_enc(std::vector<unsigned char>& buff) {

	// Chacha buffer
	ECRYPT_ctx ctx = { 0 };
	ECRYPT_keysetup(&ctx, enckey, 256, 0); // zero-length IV/nonce not used

	std::vector<unsigned char> decrypted(buff.size());
	ECRYPT_encrypt_bytes(&ctx, &buff[0], &decrypted[0], buff.size());

	// Pass to unzip function
	handle_unzip(decrypted);
}


bool crype::decrype() {

	// FIXME: it would be neat if this could work without allocating large amounts of
	//  memory at once, to protect against malicious input.
	// I know both miniz and chacha are capable of this
	// It's not a bigger priority than getting things actually working though, and it is
	//  situationally unlikely anyone other than the encryptor will be decrypting anyway

	for (;;) {
		if (feof(in) || ferror(in)) {
			break;
		}

		// read marker
		unsigned int marker = 0;
		if (fread(&marker, sizeof(unsigned int), 1, in) != 1) { // reading 1 element
			if (feof(in)) {
				break; // apparently doesn't get caught earlier
			}

			fprintf(stderr, "crype: couldn't read marker\n");
			return false;
		}

		total_chains++;

		unsigned int is_compressed = (marker & (1 << 31));
		unsigned int zlen = marker & ~(1 << 31);

		// alloc zlen memory...
		std::vector<unsigned char> buff(zlen);

		// read data from pipe
		unsigned int remaining = zlen;
		do {
			unsigned int read_size = (remaining > CRYPE_STDIN_READ_SIZE) ? CRYPE_STDIN_READ_SIZE : remaining;
			unsigned int did_read = fread(&buff[zlen - remaining], read_size, 1, in);
			if (did_read != 1) { // reading 1 element
				fprintf(stderr, "crype: couldn't read enough data\n");
				return false;
			}
			remaining -= read_size;
		} while (remaining);

		// unzip/decrypt to stdout
		if (is_compressed) {
			handle_chain_unzip_enc(buff);
		} else {
			handle_enc(buff);
		}
	}

	return true;

}
